package flac

/*
#cgo CFLAGS: -I"${SRCDIR}/../../../cgo" -march=native -Ofast -std=c99
#include "audio.h"
*/
import "C"
import (
	"bytes"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format"
	libflac "git.gammaspectra.live/S.O.N.G/goflac"
	"io"
)

type Format struct {
}

func NewFormat() Format {
	return Format{}
}

func (f Format) Open(r io.ReadSeekCloser) (audio.Source, error) {
	decoder, err := libflac.NewDecoderReader(r)
	if err != nil {
		return audio.Source{}, err
	}

	newChannel := make(chan []float32)

	go func() {
		defer close(newChannel)
		defer decoder.Close()

		frameNumber := 0

		for {
			currentFrame, err := decoder.ReadFrame()
			if err != nil {
				return
			}

			bitDepth := decoder.Depth

			if currentFrame.Depth != 0 {
				bitDepth = currentFrame.Depth
			}

			buf := make([]float32, len(currentFrame.Buffer))

			C.audio_int32_to_float32((*C.int32_t)(&currentFrame.Buffer[0]), C.size_t(len(currentFrame.Buffer)), (*C.float)(&buf[0]), C.int(bitDepth))

			newChannel <- buf

			frameNumber++
		}
	}()

	return audio.Source{
		Channels:   decoder.Channels,
		SampleRate: decoder.Rate,
		Blocks:     newChannel,
	}, nil
}

func (f Format) OpenAnalyzer(r io.ReadSeekCloser) (audio.Source, format.AnalyzerChannel, error) {
	decoder, err := libflac.NewDecoderReader(r)
	if err != nil {
		return audio.Source{}, nil, err
	}

	newChannel := make(chan []float32)
	analyzerChannel := make(format.AnalyzerChannel)

	go func() {
		defer close(newChannel)
		defer close(analyzerChannel)
		defer decoder.Close()

		frameNumber := 0

		for {
			currentFrame, err := decoder.ReadFrame()
			if err != nil {
				return
			}

			bitDepth := decoder.Depth
			channels := decoder.Channels

			if currentFrame.Depth != 0 {
				bitDepth = currentFrame.Depth
			}
			if currentFrame.Channels != 0 {
				channels = currentFrame.Channels
			}

			buf := make([]float32, len(currentFrame.Buffer))

			C.audio_int32_to_float32((*C.int32_t)(&currentFrame.Buffer[0]), C.size_t(len(currentFrame.Buffer)), (*C.float)(&buf[0]), C.int(bitDepth))

			newChannel <- buf
			analyzerChannel <- &format.AnalyzerPacket{
				Samples:    currentFrame.Buffer,
				Channels:   channels,
				BitDepth:   bitDepth,
				SampleRate: decoder.Rate,
			}

			frameNumber++
		}
	}()

	return audio.Source{
		Channels:   decoder.Channels,
		SampleRate: decoder.Rate,
		Blocks:     newChannel,
	}, analyzerChannel, nil
}

func (f Format) Encode(source audio.Source, writer io.WriteCloser, options map[string]interface{}) error {
	var bitsPerSample = 16
	var compressionLevel = 8
	var blockSize = 0
	var streamable = true

	if options != nil {
		var val interface{}
		var ok bool
		var intVal int
		var boolVal bool
		if val, ok = options["bitdepth"]; ok {
			if intVal, ok = val.(int); ok {
				bitsPerSample = intVal
			}
		}
		if val, ok = options["compression_level"]; ok {
			if intVal, ok = val.(int); ok {
				compressionLevel = intVal
			}
		}
		if val, ok = options["block_size"]; ok {
			if intVal, ok = val.(int); ok {
				blockSize = intVal
			}
		}
		if val, ok = options["streamable"]; ok {
			if boolVal, ok = val.(bool); ok {
				streamable = boolVal
			}
		}
	}

	var encoder *libflac.Encoder
	var err error

	if writeSeeker, ok := writer.(libflac.FlacWriter); ok {
		encoder, err = libflac.NewEncoderWriteSeeker(writeSeeker, source.Channels, bitsPerSample, source.SampleRate, compressionLevel, streamable, blockSize)
	} else {
		encoder, err = libflac.NewEncoderWriter(writer, source.Channels, bitsPerSample, source.SampleRate, compressionLevel, streamable, blockSize)
	}

	if err != nil {
		return err
	}

	defer encoder.Close()

	for block := range source.Blocks {
		samples := make([]int32, len(block))

		C.audio_float32_to_int32((*C.float)(&block[0]), C.size_t(len(block)), (*C.int32_t)(&samples[0]), C.int(bitsPerSample))

		err = encoder.WriteFrame(libflac.Frame{
			Rate:     source.SampleRate,
			Channels: source.Channels,
			Depth:    bitsPerSample,
			Buffer:   samples,
		})

		if err != nil {
			return err
		}
	}

	return nil
}

func (f Format) Identify(peek []byte, extension string) bool {
	return bytes.Compare(peek[:4], []byte{'f', 'L', 'a', 'C'}) == 0 || extension == "flac"
}
