package tta

/*
#cgo CFLAGS: -I"${SRCDIR}/../../../cgo" -march=native -Ofast -std=c99
#include "audio.h"
*/
import "C"
import (
	"bytes"
	"errors"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format"
	libtta "git.gammaspectra.live/S.O.N.G/go-tta"
	"io"
	"math"
	"unsafe"
)

const BlockSize = 1024 * 128

type Format struct {
}

func NewFormat() Format {
	return Format{}
}

type fakeReadWriteSeeker struct {
	r io.ReadSeekCloser
}

func (i fakeReadWriteSeeker) Read(p []byte) (n int, err error) {
	return i.r.Read(p)
}
func (i fakeReadWriteSeeker) Seek(offset int64, whence int) (int64, error) {
	return i.r.Seek(offset, whence)
}
func (i fakeReadWriteSeeker) Close() error {
	return i.r.Close()
}
func (i fakeReadWriteSeeker) Write(p []byte) (n int, err error) {
	return 0, io.ErrShortWrite
}

type fakeReadWriteSeeker2 struct {
	w io.WriteCloser
}

func (i fakeReadWriteSeeker2) Read(p []byte) (n int, err error) {
	return 0, io.ErrNoProgress
}
func (i fakeReadWriteSeeker2) Seek(offset int64, whence int) (int64, error) {
	return 0, io.ErrNoProgress
}
func (i fakeReadWriteSeeker2) Close() error {
	return i.w.Close()
}
func (i fakeReadWriteSeeker2) Write(p []byte) (n int, err error) {
	return i.w.Write(p)
}

func (f Format) Open(r io.ReadSeekCloser) (audio.Source, error) {
	decoder := libtta.MakeDecoder(fakeReadWriteSeeker{r})
	if decoder == nil {
		return audio.Source{}, errors.New("invalid decoder")
	}

	newChannel := make(chan []float32)

	info := &libtta.Info{}
	if err := decoder.GetInfo(info, 0); err != nil {
		return audio.Source{}, err
	}

	go func() {
		defer close(newChannel)

		smpSize := int(info.Nch) * ((int(info.Bps) + 7) / 8)
		buffer := make([]byte, BlockSize*smpSize)

		var writeLen int
		for {
			if writeLen = int(decoder.ProcessStream(buffer, nil)) * smpSize; writeLen == 0 {
				break
			}

			bufferSlice := buffer[:writeLen]

			bitDepth := int(info.Bps)

			nsamples := len(bufferSlice) / (bitDepth / 8)
			buf := make([]float32, nsamples)

			switch bitDepth {
			case 32:
				C.audio_int32_to_float32((*C.int32_t)(unsafe.Pointer(&bufferSlice[0])), C.size_t(nsamples), (*C.float)(&buf[0]), C.int(bitDepth))
			case 24:
				C.audio_int24_to_float32((*C.int8_t)(unsafe.Pointer(&bufferSlice[0])), C.size_t(nsamples*3), (*C.float)(&buf[0]), C.int(bitDepth))
			case 16:
				C.audio_int16_to_float32((*C.int16_t)(unsafe.Pointer(&bufferSlice[0])), C.size_t(nsamples), (*C.float)(&buf[0]), C.int(bitDepth))
			case 8:
				C.audio_int8_to_float32((*C.int8_t)(unsafe.Pointer(&bufferSlice[0])), C.size_t(nsamples), (*C.float)(&buf[0]), C.int(bitDepth))
			}

			newChannel <- buf
		}
	}()

	return audio.Source{
		Channels:   int(info.Nch),
		SampleRate: int(info.Sps),
		Blocks:     newChannel,
	}, nil
}
func (f Format) OpenAnalyzer(r io.ReadSeekCloser) (audio.Source, format.AnalyzerChannel, error) {
	decoder := libtta.MakeDecoder(fakeReadWriteSeeker{r})
	if decoder == nil {
		return audio.Source{}, nil, errors.New("invalid decoder")
	}

	newChannel := make(chan []float32)
	analyzerChannel := make(format.AnalyzerChannel)

	info := &libtta.Info{}
	if err := decoder.GetInfo(info, 0); err != nil {
		return audio.Source{}, nil, err
	}

	go func() {
		defer close(newChannel)
		defer close(analyzerChannel)

		smpSize := int(info.Nch) * ((int(info.Bps) + 7) / 8)
		buffer := make([]byte, BlockSize*smpSize)

		var writeLen int
		for {
			if writeLen = int(decoder.ProcessStream(buffer, nil)) * smpSize; writeLen == 0 {
				break
			}

			bufferSlice := buffer[:writeLen]

			bitDepth := int(info.Bps)

			nsamples := len(bufferSlice) / (bitDepth / 8)
			buf := make([]float32, nsamples)
			intBuf := make([]int32, nsamples)

			C.audio_bytes_to_int32((*C.int8_t)(unsafe.Pointer(&bufferSlice[0])), C.size_t(len(bufferSlice)), (*C.int32_t)(&intBuf[0]), C.int(bitDepth))
			C.audio_int32_to_float32((*C.int32_t)(&intBuf[0]), C.size_t(len(intBuf)), (*C.float)(&buf[0]), C.int(bitDepth))

			newChannel <- buf
			analyzerChannel <- &format.AnalyzerPacket{
				Samples:    intBuf,
				Channels:   int(info.Nch),
				BitDepth:   bitDepth,
				SampleRate: int(info.Sps),
			}
		}
	}()

	return audio.Source{
		Channels:   int(info.Nch),
		SampleRate: int(info.Sps),
		Blocks:     newChannel,
	}, analyzerChannel, nil
}

func (f Format) Encode(source audio.Source, writer io.WriteCloser, options map[string]interface{}) error {

	var bitsPerSample = 16

	if options != nil {
		var val interface{}
		var ok bool
		var intVal int
		if val, ok = options["bitdepth"]; ok {
			if intVal, ok = val.(int); ok {
				bitsPerSample = intVal
			}
		}
	}

	if bitsPerSample != 16 { //TODO
		return errors.New("not supported bits")
	}

	var encoder *libtta.Encoder

	if writeSeeker, ok := writer.(io.ReadWriteSeeker); ok {
		encoder = libtta.MakeEncoder(writeSeeker)
	} else {
		encoder = libtta.MakeEncoder(fakeReadWriteSeeker2{w: writer})
	}

	if encoder == nil {
		return errors.New("could not create encoder")
	}

	defer encoder.Close()

	err := encoder.SetInfo(&libtta.Info{
		Format:  1,
		Nch:     uint32(source.Channels),
		Bps:     uint32(bitsPerSample),
		Sps:     uint32(source.SampleRate),
		Samples: math.MaxUint32,
	})

	if err != nil {
		return err
	}

	for block := range source.Blocks {
		samples := make([]int16, len(block))

		C.audio_float32_to_int16((*C.float)(&block[0]), C.size_t(len(block)), (*C.int16_t)(&samples[0]))

		encoder.ProcessStream(unsafe.Slice((*byte)(unsafe.Pointer(&samples[0])), len(samples)*2), nil)
	}

	encoder.Finalize()

	return nil
}

func (f Format) Identify(peek []byte, extension string) bool {
	return bytes.Compare(peek[:4], []byte{'T', 'T', 'A', '1'}) == 0 || extension == "tta"
}
